/**
 * @file   ics412p3.c
 * @author Jade Cheng
 * @date   October 12, 2009
 */

#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>

static int             g_arity;         /**< The global arity input value. */
static int             g_depth;         /**< The global depth input value. */
static pthread_mutex_t g_mutex;         /**< A global lock around stdout.  */

/** Data passed to child threads. */
typedef struct arg_t {
        char         * text;            /**< The lineage text.  */
        int            depth;           /**< The current depth. */
        int            index;           /**< The sibling index. */
} arg_t;


static void spawn(const char * lineage, int depth);
static void * thread_main(void * arg);

/** The main entry point.
 *  @param argc The number of command line arguments.
 *  @param argv The command line arguments.
 *  @return EXIT_SUCCESS or EXIT_FAILURE.
 */
int main(int argc, const char ** argv)
{
        if (3 != argc) {
                fprintf(stderr, "Usage: %s <arity> <depth>\n", argv[0]);
                return EXIT_FAILURE;
        }

        if (1 != sscanf(argv[1], "%d", &g_arity) || g_arity < 1) {
                fprintf(stderr, "'%s' is an invalid arity argument\n", argv[1]);
                return EXIT_FAILURE;
        }

        if (1 != sscanf(argv[2], "%d", &g_depth) || g_depth < 1) {
                fprintf(stderr, "'%s' is an invalid depth argument\n", argv[2]);
                return EXIT_FAILURE;
        }

        /* Create child processes and wait for them. */
        pthread_mutex_init(&g_mutex, NULL);
        spawn("", 1);
        pthread_mutex_destroy(&g_mutex);
        return EXIT_SUCCESS;
}

/**
 * @param lineage The string that contains the path info of the parent process.
 * @param depth The depth of the child processes to be created.
 */
static void spawn(const char * lineage, int depth)
{
        int         i;
        pthread_t * threads;

        /* Allocate memory for child threads. */
        threads = (pthread_t *)malloc(sizeof(threads) * g_arity);

        /* Create child processes with depth one deeper and current path info. */
        for (i = 0; i < g_arity; i++) {
                arg_t     * arg;
                arg = (arg_t *)malloc(sizeof(arg_t));
                arg->depth = depth;
                arg->index = i;
                arg->text = (char *)malloc(sizeof(lineage) + 1 + 10 + 1);
                sprintf(arg->text, "%s %d", lineage, i);
                pthread_create(&threads[i], NULL, thread_main, arg);
        }

        /* Synchronize the completion of child processes. */
        for (i = 0; i < g_arity; i++)
                pthread_join(threads[i], NULL);

        /* Free the threads. */
        free(threads);
}

/**
 * @param arg The argument passed on to the child process
 * @return NULL
 */
static void * thread_main(void * param)
{
        arg_t * args = (arg_t *)param;

        /* printf is not pthread safe, so lock before entering shared buffer. */
        pthread_mutex_lock(&g_mutex);
        printf("Lineage =%s%s\n", args->text, args->depth == g_depth ? " *" : "");
        pthread_mutex_unlock(&g_mutex);

        /* If current process is not the leaf continue sprawning. */
        if (args->depth != g_depth)
                spawn(args->text, args->depth + 1);

        /* Free the memory for related objects. */
        free(args->text);
        free(args);
        return NULL;
}
